#include "aadl_dispatch.h"
#include "aadl_ports.h"
#include "aadl_multiarch.h"
#include "aadl_error.h"

#include "deployment.h"

#ifdef USE_POSIX

#ifdef DEBUG
#include <stdio.h>
#endif // DEBUG

#endif // USE_POSIX

#if defined DEBUG && defined USE_OSEK
#include "ecrobot_interface.h"
#endif


StatusType Stop_Thread(thread_config_t * config)
{
#if defined(RUNTIME_DEBUG) && (defined(USE_POSIX) || defined(USE_POK))
  printf("Stopping thread %s\n",config->name);
#endif
	config->state=STOPPED_THREAD_STATE;
	return stop_thread_multiarch(config);
}

StatusType Create_Thread(thread_config_t * config)
{
	return create_thread_multiarch(config);
}

StatusType Start_Thread(thread_config_t * config)
{
#if defined(RUNTIME_DEBUG) && (defined(USE_POSIX) || defined(USE_POK))
  printf("Starting thread %s\n",config->name);
#endif

  int ret = init_thread_config(config);
  if(ret)
    return ret;

  switch (config->protocol)
  {
    case PERIODIC:
    {
      periodic_thread_config_t * per_thr = config->thread;
      per_thr->parent = config;
      init_periodic_config(per_thr);
      break;
    }
    case SPORADIC:
    {
      sporadic_thread_config_t * spo_thr = config->thread;
      spo_thr->parent = config;
      init_sporadic_config(spo_thr);
      break;
    }
    case APERIODIC:
    {
      aperiodic_thread_config_t * aper_thr = config->thread;
      aper_thr->parent = config;
      init_aperiodic_config(aper_thr);
      break;
    }
    case HYBRID:
    {
      hybrid_thread_config_t * hyb_thr = config->thread;
      hyb_thr->parent = config;
      init_hybrid_config(hyb_thr);
      break;
    }
    case TIMED:
    {
      timed_thread_config_t * tim_thr = config->thread;
      tim_thr->parent = config;
      init_timed_config(tim_thr);
      break;
    }
    case TIME_TRIGGERED:
    {
      time_triggered_thread_config_t * tt_config = config->thread;
      tt_config->parent = config;
      init_time_triggered_config(tt_config);
      break;
    }
    default:
      return -1;
  }

  config->state = INITIALIZING_THREAD_STATE;
  return start_thread_multiarch(config);
}


int init_thread_config(thread_config_t * config)
{
  int ret = init_thread_config_multiarch(config);
  return ret;
}

int init_sporadic_config (sporadic_thread_config_t * info)
{
  int ret = init_port_list_mutex(info->parent->global_q);
  return ret;
}

int init_aperiodic_config (aperiodic_thread_config_t * info)
{
  int ret = init_port_list_mutex(info->parent->global_q);
  return ret;
}

int init_periodic_config (periodic_thread_config_t * info)
{
  int ret = E_OK;

  if (info->parent->global_q != NULL)
    ret = init_port_list_mutex(info->parent->global_q);
  ret = init_periodic_config_multiarch(info);
  return ret;
}

int init_hybrid_config (hybrid_thread_config_t * info)
{
  int ret = E_OK;
  if (info->parent->global_q != NULL)
    init_port_list_mutex(info->parent->global_q);
  init_hybrid_timer(info);
  return ret;
}

int init_timed_config (timed_thread_config_t * info)
{
  int ret = init_port_list_mutex(info->parent->global_q);
  return ret;
}

// implementation if init_time_triggered_config in multiarch
// int init_time_triggered_config (time_triggered_thread_config_t * info);

int Init_Config (thread_config_t * config,
		error_code_t (*recovery_entrypoint)(thread_config_t * config, error_code_t error),
		void * (*compute_entrypoint)(),
		unsigned int id)
{
  config->recovery_entrypoint = recovery_entrypoint;
  config->compute_entrypoint = compute_entrypoint;
  config->id = id;
  if (config->global_q != NULL)
  {
    config->global_q->parent = config;
  }

  return 0;
}


StatusType await_periodic_dispatch(periodic_thread_config_t * info)
{
  StatusType ret = await_periodic_dispatch_multiarch(info);
  return ret;
}

StatusType await_time_triggered_dispatch(time_triggered_thread_config_t * config)
{
  StatusType ret = await_time_triggered_dispatch_multiarch(config);
  return ret;
}

StatusType await_sporadic_dispatch(sporadic_thread_config_t * info)
{
  StatusType status = E_OK;

  // wait next period
  if(info->timing_wait_phase)
  {
    sporadic_thread_sleep(info);
    info->timing_wait_phase = 0;
  }
  info->timing_wait_phase = 1;

  // wait for messages
  status = lock_port_list(info->parent->global_q);

  if (status != E_OK)
    return status;

  if(info->parent->global_q->msg_nb_at_dispatch>=info->parent->global_q->msg_nb)
  {
    info->parent->global_q->waiting = info->parent->global_q->waiting+1;
    status = wait_port_list(info->parent->global_q);
    info->parent->global_q->waiting = info->parent->global_q->waiting-1;
  }

  status = unlock_port_list(info->parent->global_q);
  
  return status;
}

StatusType await_aperiodic_dispatch(aperiodic_thread_config_t * info)
{
  StatusType status = lock_port_list(info->parent->global_q);
  if (status != E_OK)
    return status;


  if(info->parent->global_q->msg_nb_at_dispatch>=info->parent->global_q->msg_nb)
  {
	  info->parent->global_q->waiting += 1;
	  status = wait_port_list(info->parent->global_q);// FIXME manage error
	  info->parent->global_q->waiting -= 1;
  }
  status = unlock_port_list(info->parent->global_q);

  return status;
}

StatusType await_hybrid_dispatch(hybrid_thread_config_t * info)
{
  StatusType status = lock_port_list(info->parent->global_q);
  if (status != E_OK)
    return status;

  if(info->parent->global_q->msg_nb_at_dispatch>=info->parent->global_q->msg_nb)
  {
    info->parent->global_q->waiting += 1;
    status = timedwait_hybrid(info);
    info->parent->global_q->waiting -= 1;

    // If it's not an event which woke up the hybrid thread
    if(info->parent->global_q->msg_nb_at_dispatch>=info->parent->global_q->msg_nb)
      hybrid_reset_timer(info);

    // FIXME Need to call ClearEvent for osek
  }

  status = unlock_port_list(info->parent->global_q);
  return status;
}

StatusType await_timed_dispatch(timed_thread_config_t * info)
{
  StatusType status;
  lock_port_list(info->parent->global_q);

  if(info->parent->global_q->msg_nb_at_dispatch>=info->parent->global_q->msg_nb)
  {
    info->parent->global_q->waiting += 1;
    status = timedwait_timed_thread(info);
    info->parent->global_q->waiting -= 1;
  }

  unlock_port_list(info->parent->global_q);
  return status;
}

unsigned int max_exec_time[TASKS_NB];
extern process_config_t this_process;

void init_iteration_counters(thread_config_t * config, char init);

StatusType Await_Dispatch(thread_config_t * config)
{
  StatusType ret = E_OK;
  // reset freshness of input data ports
  int port_iter;
  for(port_iter=0; port_iter<config->in_ports_size;port_iter++) {
    port_reference_t * port_ref = config->in_ports[port_iter];
    if(port_ref->type != DATA)
      continue;
    thread_input_port_t * in_port = (thread_input_port_t *) port_ref->directed_port;
    in_port->fresh = 0;
  }
  // if not in mode, call await_mode
  // may occur if mode changed while thread was suspended
  if(!Is_In_Current_Mode(config)) {
    Await_Mode(config);
    init_iteration_counters(config, 1);
  }
  else
  {
	config->state = SUSPENDED_THREAD_STATE;
	runtime_time_t finish_t;
	get_time_multiarch(&(config->finish_time));
#if defined BENCHMARK && (defined USE_POSIX || defined USE_POK)
	unsigned int exec_time = finish_t.msec-(config->dispatch_time.msec+((config->dispatch_time.nsec/1000000)*1000000));
	if(exec_time>max_exec_time[config->id])
	    max_exec_time[config->id] = exec_time;
	printf("Thread: %s\n", config->name);
	printf("Max Response Time (ms) = %u\n", max_exec_time[config->id]);
#endif
#if defined BENCHMARK && defined USE_OSEK
	unsigned int exec_time = finish_t.msec-(config->dispatch_time.msec+((config->dispatch_time.nsec/1000000)*1000000));
	display_goto_xy(0, config->id*2);
	if(strlen(config->name)<15)
		display_string(config->name);
	else
	{
		char txt1[16];
		int i, k=0;

		for(i=strlen(config->name)-10;i<=strlen(config->name);i++)
		{
			txt1[k]=config->name[i];
			k++;
		}
		display_string(txt1);
	}


    char txt2[11] = "RT (ms) = ";
    char txt3[5];
    if(finish_t.msec > 3000 && exec_time>max_exec_time[config->id])
    	max_exec_time[config->id] = exec_time;

    itoa(max_exec_time[config->id], txt3, 10);
    display_goto_xy(0, config->id*2+1);
    display_string(strcat(txt2,txt3));
    display_update();

#endif
    switch (config->protocol)
    {
      case PERIODIC:
        await_periodic_dispatch(config->thread);
        break;
      case SPORADIC:
    	  await_sporadic_dispatch(config->thread);
    	  break;
      case APERIODIC:
        await_aperiodic_dispatch(config->thread);
        break;
      case HYBRID:
        await_hybrid_dispatch(config->thread);
        break;
      case TIMED:
        await_timed_dispatch(config->thread);
        break;
      case TIME_TRIGGERED:
    	await_time_triggered_dispatch(config->thread);
    	break;
      default:
        return -1;
    }
    get_time_multiarch(&(config->dispatch_time));

    // if not in mode, call await_mode
    // may occur if mode changed while thread was suspended
    if(!Is_In_Current_Mode(config)) {
    	Await_Mode(config);
	init_iteration_counters(config, 1);
    }
  }

#if defined(RUNTIME_DEBUG) && (defined(USE_POSIX) || defined(USE_POK))
  runtime_time_t dispatch_t;
  get_time_multiarch(&dispatch_t);
  printf("Dispatch %s @t=%d\n", config->name, dispatch_t.msec);
#endif

  // check inputs with immediate connections are available
  int i;
  for(i=0; i<config->in_ports_size; i++)
  {
	  port_reference_t * port_ref = config->in_ports[i];
	  thread_input_port_t * in_port = (thread_input_port_t*) port_ref->directed_port;

	  if(!in_port->in_immediate_connection)
		  continue;
	  
	  int fresh;
	  Updated(port_ref, &fresh);
	  if(!fresh)
	  {
#if defined DEBUG && (defined USE_POK || defined USE_POSIX)
		  printf("released task but port content is not fresh!\n");
#endif
		  error_code_t err = Error_Handler(MISSING_INPUT, in_port->parent->parent, NULL);
		  if(err)
			  return err;
	  }
  }
  config->state = RUNNING_THREAD_STATE;
  return ret;
}

StatusType keep_busy_until(unsigned int finish_time_ns)
{
  runtime_time_t current_time;

#if defined(RUNTIME_DEBUG) && (defined(USE_POSIX)||defined(USE_POK))
  get_time_multiarch(&current_time);
  printf("Busy from %d\n", current_time.msec*1000000+current_time.nsec);
#endif
  char reached = 0;
  while(!reached)
    {
	  get_time_multiarch(&current_time);
      if(current_time.msec*1000000+current_time.nsec >= finish_time_ns)
    	  reached = 1;
    }
#if defined(RUNTIME_DEBUG) && (defined(USE_POSIX)||defined(USE_POK))
  get_time_multiarch(&current_time);
  printf("Busy until %d\n", current_time.msec*1000000+current_time.nsec);
#endif
  return E_OK;
}
