#include "aadl_ports.h"
#include "aadl_multiarch.h"

StatusType Put_Value(port_reference_t * port, void * value)
{
  StatusType status = E_OK;

  port_direction_t direction = port->direction;
  if(direction!=OUT){
	status = INVALID_PARAMETER;
  }

  thread_output_port_t * out_port = (thread_output_port_t*) port->directed_port;
  out_port->updated = 1;
  if(value != NULL)
    memcpy(out_port->port_variable, value, port->msg_size);

  return status;
}

StatusType Send_Output(port_reference_t * src_port)
{
  StatusType status = E_OK;
  port_direction_t direction = src_port->direction;
  if(direction!=OUT && direction!=INOUT) {
    status = INVALID_PARAMETER;
#if defined DEBUG && defined USE_POSIX
          printf("Send_Output: ERROR, parameter is not output port\n");
#endif
  }

  component_t parent_type = src_port->parent_type;
  if(parent_type!=THREAD) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && defined USE_POSIX
	  printf("Send_Output: ERROR, parameter is not a thread port\n");
#endif
  }

  thread_output_port_t * out_port = (thread_output_port_t*) src_port->directed_port;
  if(out_port->updated==0)
    return status;
  
  Do_Send_Output(src_port);

  out_port->updated=0;
  return status;
}

/* Note: Is_Empty cannot be replaced by Get_Count since
 * Is_Empty is supposed to be used in a critical section
 * Get_Count implements its own critical section
 */
char Is_Empty(port_reference_t * port)
{
	char is_empty = 0;
	thread_input_port_t * in_port = (thread_input_port_t*) port->directed_port;
	if (port->type == EVENT_DATA)
	  {
	    event_data_port_t * port_ed = in_port->content;
	    if(port_ed->beg_read_idx == port_ed->end_read_idx)
	    {
	    	is_empty = 1;
	    }
    }
	else if (port->type == EVENT)
	  {
	    event_port_t * port_e = in_port->content;
	    if(port_e->beg_read_idx == port_e->end_read_idx)
	    {
	    	is_empty = 1;
	    }
    }
	return is_empty;
}

StatusType Next_Value(port_reference_t * dst_port, void * value)
{
  StatusType status = E_OK;

  port_direction_t direction = dst_port->direction;
  if(direction!=IN) {
    status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
          printf("Next_Value: ERROR, parameter is not an input port\n");
#endif
  }

  component_t parent_type = dst_port->parent_type;
  if(parent_type!=THREAD) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	  printf("Next_Value: ERROR, parameter is not a thread port\n");
#endif
  }

  thread_input_port_t * in_port = (thread_input_port_t*) dst_port->directed_port;

  status = lock_port_reference(in_port);

  if (status != E_OK)
  {
    Error_Handler(status, in_port->parent->parent, NULL);
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
    printf("Next value: ERROR %d when acquiring lock\n", status);
#endif
    return status;
  }


  if(Is_Empty(dst_port))
  {
	  status = E_OS_LIMIT;
	  unlock_port_reference(in_port);
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
      printf("Next value: ERROR, empty queue\n");
#endif
  }
  if(status!=E_OK)
  {
	  Error_Handler(status, in_port->parent->parent, NULL);
	  return status;
  }
  if (dst_port->type == EVENT_DATA)
  {
    event_data_port_t * port_ed = in_port->content;
    port_ed->beg_read_idx = (port_ed->beg_read_idx+1)%port_ed->queue_size;

#if (defined USE_POSIX || defined USE_POK) && defined RUNTIME_DEBUG
    printf("Next value: first index is %d, %d\n", port_ed->beg_read_idx);
#endif
  }
  else if (dst_port->type == EVENT)
  {
    event_port_t * port_event = in_port->content;
    port_event->beg_read_idx=(port_event->beg_read_idx+1)%port_event->queue_size;

#if (defined USE_POSIX || defined USE_POK) && defined RUNTIME_DEBUG
    printf("Next value: first index is %d, %d\n", port_event->beg_read_idx);
#endif
  }

  if(Is_Empty(dst_port))
  {
	  status = E_OS_LIMIT;
	  unlock_port_reference(in_port);
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
      printf("Next value: ERROR, empty queue\n");
#endif
  }
  if(status!=E_OK)
  {
	  Error_Handler(status, in_port->parent->parent, NULL);
	  return status;
  }

  in_port->parent->msg_nb--;

  if (dst_port->type == DATA)
  {
	mode_id_t current_mode;
	Current_System_Mode(&current_mode);
    data_port_t * port_data = in_port->content;
    memcpy(value, port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].reading_data, dst_port->msg_size);
  }
  else if (dst_port->type == EVENT_DATA)
  {
    event_data_port_t * port_ed = in_port->content;
    char * buf = (char*) port_ed->offset;
    memcpy(value,
           &buf[(port_ed->beg_read_idx+1)*(dst_port->msg_size/sizeof(char))],
		   dst_port->msg_size);
  }
  else
  {
    event_port_t * port_event = in_port->content;
    char * buf = (char*) port_event->offset;
    memcpy(value,
           &buf[port_event->beg_read_idx+1],
		   dst_port->msg_size);
  }
  status = unlock_port_reference(in_port);
  if (status != E_OK)
    Error_Handler(status, in_port->parent->parent, NULL);

  return status;
}

StatusType Get_Value(port_reference_t * dst_port, void * dst)
{
  StatusType status = E_OK;

  port_direction_t direction = dst_port->direction;
  if(direction!=IN) {
    status = INVALID_PARAMETER;
#if defined DEBUG && defined USE_POSIX
          printf("Get_Value: ERROR, parameter is not an input port\n");
#endif
  }

  component_t parent_type = dst_port->parent_type;
  if(parent_type!=THREAD) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	  printf("Get_Value: ERROR, parameter is not a thread port\n");
#endif
  }

  thread_input_port_t * in_port = (thread_input_port_t*) dst_port->directed_port;

  lock_port_reference(in_port);

  if(Is_Empty(dst_port))
  {
	  status = E_OS_LIMIT;
	  unlock_port_reference(in_port);
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
      printf("Get_Value: ERROR, empty queue\n");
#endif
  }
  if(status!=E_OK)
  {
	  Error_Handler(status, in_port->parent->parent, NULL);
	  return status;
  }

  if (dst_port->type == DATA)
  {
    data_port_t * port_data = in_port->content;
    mode_id_t current_mode;
	Current_System_Mode(&current_mode);
    void * addr = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].reading_data;
    memcpy(dst, addr, dst_port->msg_size);
  }
  else if (dst_port->type == EVENT_DATA)
  {
    event_data_port_t * port_ed = in_port->content;
    char * buf = (char*) port_ed->offset;
/*    int i=0;
    printf("queue content: ");
    for(i=0;i<port_ed->queue_size;i++)
    {
    	printf("%d ", buf[i*(dst_port->msg_size/sizeof(char))]);
    }
    printf("\n");
    printf("get value @index = %d\n", (port_ed->beg_read_idx+1)); */

    memcpy(dst,
    		&buf[(port_ed->beg_read_idx)*(dst_port->msg_size/sizeof(char))],
			dst_port->msg_size); //No more read at beg_read_idx + 1
  }
  else
  {
    event_port_t * port_event = in_port->content;
    char * buf = (char*) port_event->offset;
    memcpy(dst,
           &buf[port_event->beg_read_idx+1],
		   dst_port->msg_size);
  }

  unlock_port_reference(in_port);

  return status;
}

StatusType Receive_Input(port_reference_t * dst_port)
{
  StatusType status = E_OK;

  port_direction_t direction = dst_port->direction;
  if(direction!=IN) {
    return INVALID_PARAMETER;
  }

  component_t parent_type = dst_port->parent_type;
  if(parent_type!=THREAD) {
	  return INVALID_PARAMETER;
  }

  thread_input_port_t * in_port = dst_port->directed_port;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
	printf("Receive_Input connection numbers = %d\n", in_port->connected_ports_length);
#endif

  int i;
  for(i=0; i<in_port->connected_ports_length; i++) {
    port_connection_t * cnx = in_port->sources[i];
    status = Receive_Input_Source(cnx);
    if(status!=E_OK)
  	  return status;
  }

  in_port->parent->msg_nb_at_dispatch = in_port->parent->msg_nb;

  return status;
}

StatusType Get_Count(port_reference_t * port, int * count_res)
{
  StatusType status = E_OK;

  port_direction_t direction = port->direction;
  if(direction!=IN) {
    status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
          printf("Get_Count: ERROR, parameter is not an input port\n");
#endif
  }

  component_t parent_type = port->parent_type;
  if(parent_type!=THREAD) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	  printf("Get_Count: ERROR, parameter is not a thread port\n");
#endif
  }

  thread_input_port_t * in_port = (thread_input_port_t*) port->directed_port;

  status = lock_port_reference(in_port);
  if (status != E_OK)
  {
    Error_Handler(status, in_port->parent->parent, NULL);
    return status;
  }

  if (port->type == DATA)
  {
    *count_res = in_port->fresh;
  }
  else if (port->type == EVENT_DATA)
  {
    event_data_port_t * port_ed = in_port->content;
    if(port_ed->end_read_idx < port_ed->beg_read_idx)
      *count_res = port_ed->queue_size-port_ed->beg_read_idx+port_ed->end_read_idx;
    else
      *count_res = port_ed->end_read_idx - port_ed->beg_read_idx;
  }
  else
  {
    event_port_t * port_event = in_port->content;
    if(port_event->end_read_idx < port_event->beg_read_idx)
      *count_res = port_event->queue_size-port_event->beg_read_idx+port_event->end_read_idx;
    else
      *count_res = port_event->end_read_idx - port_event->beg_read_idx;
  }

  status = unlock_port_reference(in_port);
  if (status != E_OK)
    Error_Handler(status, in_port->parent->parent, NULL);
#if defined RUNTIME_DEBUG && (defined USE_POK || defined USE_POSIX)
  printf("Get_Count result on port = %d, on thread %s\n", *count_res, in_port->parent->parent->name);
#endif
  return status;
}

StatusType Updated(port_reference_t * port, int * fresh_flag)
{
  StatusType status = E_OK;
  port_direction_t direction = port->direction;
  if(direction!=IN) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	  printf("Updated: ERROR, parameter is not an input port\n");
#endif
  }

  component_t parent_type = port->parent_type;
  if(parent_type!=THREAD) {
	  status = INVALID_PARAMETER;
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	  printf("Updated: ERROR, parameter is not a thread port\n");
#endif
  }

  thread_input_port_t * in_port = (thread_input_port_t*) port->directed_port;

  status = lock_port_list(in_port->parent);
  *fresh_flag = in_port->fresh;
  status = unlock_port_list(in_port->parent);

  return status;
}

