#ifndef __AADL_TYPES_H_
#define __AADL_TYPES_H_

#ifdef USE_TPL_OS 

#include "tpl_os_types.h"
#include "tpl_os_resource.h"
#include "tpl_os_event.h"

#else

#ifdef USE_POSIX

#include <signal.h>
#include <string.h>
#include <pthread.h>

#define E_OK 0
#define E_OS_LIMIT -100;

typedef int StatusType;
struct periodic_thread_config;

#else

#ifdef USE_OSEK

#include "kernel.h"
#define NULL 0

#else

#ifdef USE_POK

#define E_OK 0
#define E_OS_LIMIT -100;

#ifdef POK_NEEDS_DEBUG
#define DEBUG
#endif

#include <arinc653/types.h>
typedef RETURN_CODE_TYPE StatusType;

#include <arinc653/semaphore.h>
#include <arinc653/event.h>

#endif // USE_POK

#endif // USE_OSEK

#endif // USE_POSIX

#endif // USE_TPL_OS


#include "dimensions.h"

#if MODES_NB == 1
typedef int mode_id_t;
#endif

typedef struct runtime_time_t
{
	unsigned long msec;
	unsigned long nsec;
} runtime_time_t;

typedef enum error_code_t
{
  OK,
  EMPTY_QUEUE,
  FULL_QUEUE,
  LOCK_ERROR,
  OUT_OF_BOUND,
  INVALID_PARAMETER,
  INVALID_SERVICE_CALL,
  MISSING_INPUT,
  INIT_HYPERPERIOD
} error_code_t;

typedef enum access_protocol_t {
  NONE = 1,
  LOCK = 2,
  ICPP = 3
} access_protocol_t;

typedef struct resource_t {
  access_protocol_t protocol;
  void * protocol_config; // can be NULL, lock_access_protocol_t or icpp_access_protocol_t
} resource_t;

typedef struct lock_access_protocol_t {
#ifdef USE_POSIX
  pthread_mutex_t * rez;
#endif
#ifdef USE_OSEK
  ResourceType * rez;
#endif
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SEMAPHORE_ID_TYPE * rez;
#endif
} lock_access_protocol_t;

typedef struct icpp_access_protocol_t {
  unsigned int ceiling_priority;
  lock_access_protocol_t * lock;
  unsigned int base_priority; // set dynamically when calling Get_Resource
} icpp_access_protocol_t;


/* Queue id type */
typedef int port_id_t;

typedef enum port_type_t {
  DATA,
  EVENT,
  EVENT_DATA
} port_type_t;

typedef enum port_direction_t {
  IN = 1,
  OUT = 2,
  INOUT = 3
} port_direction_t;

typedef struct event_port_t {
  int queue_size;    // max size of the queue
  int beg_read_idx;  // Begining of reading
  int end_read_idx;  // End of reading
  int end_write_idx; // End of writing
  void * offset;
} event_port_t;

typedef struct event_data_port_t {
  int queue_size;    // max size of the queue
  int beg_read_idx;  // Begining of reading
  int end_read_idx;  // End of reading
  int end_write_idx; // End of writing
  void * offset;
} event_data_port_t;

typedef struct double_buffer_t {
	void * reading_data;
	void * writing_data;
} double_buffer_t;

typedef struct data_port_t {
  unsigned char nb_mode;
  int * src_index_by_mode; // array of indexes in [reading/writing]_data_per_mode
  double_buffer_t * data_rw_per_mode;
} data_port_t;

struct port_list_t;

typedef enum component_t {
	THREAD,
	PROCESS
} component_t;

struct connection_protocol_t;

typedef struct port_reference_t {
	port_type_t type;
	port_direction_t direction;
	component_t parent_type;
	void * parent; // can be thread_config_t or process_config_t
	void * directed_port; 	// can be thread_output_port_t or thread_input_port_t
							// or process_port_t or processor_port_t or bus_port_t
	unsigned int msg_size;
} port_reference_t;

typedef port_reference_t* port_reference_addr;

typedef struct port_connection_t {
	port_reference_addr source;
	port_reference_addr destination;
	struct connection_protocol_t * protocol;
} port_connection_t;

typedef port_connection_t * port_connection_addr;

typedef struct process_port_t {
	unsigned int identifier;
	void * sys_port;
} process_port_t;

typedef process_port_t* process_port_addr;

typedef struct process_config_t
{
  struct thread_config_t * threads[TASKS_NB];
  unsigned int criticality_levels_nb;
  unsigned int criticality[MODES_NB];
  unsigned int hyperperiod;
} process_config_t;

typedef struct thread_input_port_t {
  struct port_list_t * parent;
  port_reference_t * port_ref;

  void * content; // can be event_port_t or data_port_t or event_data_port_t

  int connected_ports_length; // number of sources of a dst port

  port_connection_addr * sources;

  char fresh;
  char in_immediate_connection;
} thread_input_port_t;

typedef struct thread_output_port_t {
  void * port_variable;
  char updated;

  /*
     * destinations contains references of ports connected to self
     *
     * from index 0 to index connected_ports_length-1 we have references
     * to ports connected to self in all modes
     *
     * from index connected_ports_length to index
     * connected_ports_length+dest_index_by_mode[nb_mode]
     * we have references to ports connected to self in each mode
     *
     * To access connected ports in mode m:
     *
     * for(j=dest_index_by_mode[m]; j<dest_index_by_mode[m+1];j++)
     *   destinations[j]
     *
     * */

  unsigned int nb_mode;
  int connected_ports_length; // number of destinations of a src port
  int * dest_index_by_mode; // array of indexes in destinations
  port_connection_addr * destinations;
} thread_output_port_t;

typedef struct data_access_reference_t {
  void * data;
  resource_t * resource;
  int size;
} data_access_reference_t;

typedef data_access_reference_t* data_access_reference_addr;

struct thread_config_t;

typedef struct port_list_t {
  struct thread_config_t * parent;
  char waiting;
  unsigned int msg_nb;
  unsigned int msg_nb_at_dispatch;
  int ports_nb;
  port_reference_t ** port_references;

#ifdef USE_POSIX // TODO: replace type of rez by resource_t
  pthread_mutex_t * rez;
  pthread_cond_t * event;
#endif
#ifdef USE_OSEK
  ResourceType * rez;
  EventMaskType * event;
#endif
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SEMAPHORE_ID_TYPE * rez;
  SEMAPHORE_ID_TYPE * event;
#endif

} port_list_t;


typedef struct wd_desc {
  int wd_alrm;
#if defined USE_POSIX
#ifndef USE_MACH 
  timer_t timer_id;
  struct itimerspec tDef;
  struct periodic_thread_config *task_descriptor; 
#endif
#endif
} wd_desc_t; 

typedef enum dispatch_protocol_t {
  PERIODIC,
  SPORADIC,
  HYBRID,
  APERIODIC,
  BACKGROUND,
  TIMED,
  TIME_TRIGGERED
} dispatch_protocol_t;


typedef enum thread_state_t
{
  STOPPED_THREAD_STATE,
  INITIALIZING_THREAD_STATE,
  AWAITING_MODE_THREAD_STATE,
  SUSPENDED_THREAD_STATE,
  RUNNING_THREAD_STATE
} thread_state_t;


typedef struct m_k_firm_t
{
	unsigned int tfe_count;
	unsigned const int tfe_limit; // m in m-k firm
	unsigned const int tfe_count_frame; // k in m-k firm
	unsigned char * tfe_history; // array of size k
	unsigned int tfe_history_idx;
} m_k_firm_t;

typedef struct thread_config_t
{
  // TODO: add state and update it in runtime services
  dispatch_protocol_t protocol;
  port_list_t * global_q;
  void* thread;	  // can be periodic_thread_config_t or sporadic_thread_config_t
  	  	  	  	  // or timed_thread_config_t or hybride_thread_config_t
  	  	  	  	  // or time_triggered_thread_config_t
  const unsigned int priority;
  const unsigned int deadline;

  unsigned int criticality_levels_nb;
  unsigned int criticality[MODES_NB];
  void * thread_id; // can be NULL or pthread_t (POSIX), or TaskType (OSEK)...
// fields for waiting mode
#ifdef USE_POSIX
  pthread_cond_t * wait_mode_event;
#endif
#ifdef USE_OSEK
  EventMaskType * wait_mode_event;
#endif
#ifdef USE_POK
  SEMAPHORE_ID_TYPE * wait_mode_event;
#endif
  int out_ports_size;
  port_reference_t ** out_ports;
  int in_ports_size;
  port_reference_t ** in_ports;

  m_k_firm_t * m_k_constraint;

  thread_state_t state;
  unsigned int id;
  error_code_t (*recovery_entrypoint)(struct thread_config_t * config, error_code_t error);
  void (*compute_entrypoint)();
  char * name;

  runtime_time_t dispatch_time;
  runtime_time_t finish_time;

} thread_config_t;

typedef thread_config_t * thread_config_addr;
typedef thread_config_addr * thread_config_addr_array;

typedef struct periodic_thread_config
{
  const unsigned int period;
  const unsigned int offset;
  unsigned long iteration_counter;
#if defined USE_MACH
  pthread_mutex_t * time_interval_mutex;
  pthread_cond_t * time_interval_cond;
  struct timespec * start_timespec;
#elif defined USE_OSEK
  AlarmType periodic_alarm; // has to be cyclic alarm and call the task (if it doesn't create another thread)
#endif
  wd_desc_t  wd;
  thread_config_t * parent;
} periodic_thread_config_t;

typedef struct time_triggered_thread_config
{
#ifdef USE_POSIX
  pthread_mutex_t * release_rez;
  pthread_cond_t * release_event;
#elif defined USE_OSEK
  EventMaskType * release_event;
#elif defined USE_POK
  SEMAPHORE_ID_TYPE * release_event;
#endif
  thread_config_t * parent;
  char released;
} time_triggered_thread_config_t;

typedef struct sporadic_thread_config
{
  const unsigned int time_interval;
  char timing_wait_phase;
  thread_config_t * parent;

} sporadic_thread_config_t;

typedef struct aperiodic_thread_config_t
{
  thread_config_t * parent;
} aperiodic_thread_config_t;

typedef struct hybrid_thread_config_t
{
  const unsigned int time_interval;
#if defined USE_POSIX
  unsigned int nb_period;
  struct timespec * start_timespec;
#elif defined USE_OSEK
  AlarmType time_interval_alarm; // This alarm doesn't reactivate the task, but just sends an event on the port_list
#endif
  thread_config_t * parent;

} hybrid_thread_config_t;

typedef struct timed_thread_config_t
{
#ifdef USE_POK
  SYSTEM_TIME_TYPE time_interval;
#else
  const unsigned int time_interval;
#endif
  thread_config_t * parent;
} timed_thread_config_t;

typedef enum mode_switch_protocol_t
{
	EMERGENCY_MODE_SWITCH,
	PLANNED_MODE_SWITCH
} mode_switch_protocol_t;

typedef struct time_triggered_schedule_event
{
  unsigned int sched_time;
  // tasks to release
  unsigned int task_release_nb;
  unsigned int task_release_idx[TASKS_NB];
  // tasks deadline
  unsigned int task_deadline_nb;
  unsigned int task_deadline_idx[TASKS_NB];
  // tasks placement
  unsigned int task_placement_nb; // note: initialize to tasks_nb at time 0
  unsigned int task_placement_idx[TASKS_NB]; // note: used to bind all threads at time 0
  unsigned int placement_core[CORES_NB];
  // tasks promotion
  unsigned int task_promotion_nb;
  unsigned int task_promotion_idx[TASKS_NB];
  // tasks degradation
  unsigned int task_degradation_nb;
  unsigned int task_degradation_idx[TASKS_NB];
} time_triggered_schedule_event_t;

typedef time_triggered_schedule_event_t * time_triggered_schedule_event_addr_t;

typedef enum connection_protocol_type_t {
	SAMPLED,
	IMMEDIATE,
	PDP_LOWMFP,
	PDP_LOWET,
	PDP_LOWMC
} connection_protocol_type_t;

typedef struct connection_protocol_t
{
	connection_protocol_type_t protocol_type;
	void * protocol_attr; // can be PDP_lowET_attr_t, PDP_lowMC_attr_t
} connection_protocol_t;

typedef connection_protocol_t * connection_protocol_addr;

typedef struct PDP_lowMFP_attr_t
{
	unsigned int read_lookup_table_size;
	unsigned int write_lookup_table_size;
} PDP_lowMFP_attr_t;

typedef struct PDP_lowET_attr_t
{
	unsigned int * read_lookup_table;
	unsigned int read_lookup_table_size;
	unsigned int * write_lookup_table;
	unsigned int write_lookup_table_size;
} PDP_lowET_attr_t;

typedef PDP_lowET_attr_t * PDP_lowET_attr_addr;

typedef struct PDP_lowMC_attr_t
{

} PDP_lowMC_attr_t;

StatusType Put_Value(port_reference_t * port, void * value);

StatusType Send_Output(port_reference_t * port);

StatusType Next_Value(port_reference_t * port, void * dst);

StatusType Get_Value(port_reference_t * port, void * dst);

StatusType Receive_Input(port_reference_t * port);

StatusType Get_Count(port_reference_t * port, int * count_res);

StatusType Updated(port_reference_t * port, int * fresh_flag);

StatusType Await_Mode(thread_config_t * config);

StatusType Await_Dispatch(thread_config_t * global_q);


#endif
