#define _GNU_SOURCE 1
#include "aadl_multiarch.h"
#include "aadl_modes_standard.h"
#include <pthread.h>
#include <time.h>
#include <sched.h>
#include <limits.h>
#include <unistd.h>

struct timespec start_timespec;

void get_time_multiarch(runtime_time_t * time)
{
  struct timespec current_time;
  clock_gettime(CLOCK_MONOTONIC, &current_time);
  time->msec = current_time.tv_sec*1000 - start_timespec.tv_sec*1000;
  time->nsec = current_time.tv_nsec - start_timespec.tv_nsec;
}

void get_start_time(runtime_time_t * time)
{
	time->msec = start_timespec.tv_sec*1000+start_timespec.tv_nsec/1000000;
	time->nsec = start_timespec.tv_nsec-(start_timespec.tv_nsec/1000000)*1000000;
}

void set_start_time_multiarch()
{

  clock_gettime(CLOCK_MONOTONIC, &start_timespec);
}

StatusType set_thread_priority_multiarch(thread_config_t * config, unsigned int priority)
{
	pthread_t * tid = (pthread_t*) config->thread_id;
	return pthread_setschedprio(*tid, priority);
}

int init_thread_config_multiarch(thread_config_t * config)
{
  int ret = pthread_cond_init(config->wait_mode_event, NULL);
  return ret;
}

int init_periodic_config_multiarch(periodic_thread_config_t * config)
{
  int ret = 0;
  config->iteration_counter=0;
  return ret;
}

extern pthread_mutex_t init_barrier_mutex;
extern pthread_cond_t init_barrier_cond;
extern unsigned long reached_init_barrier;
extern mode_id_t current_mode;

StatusType Await_Mode_multiarch(thread_config_t * config)
{
  StatusType ret = E_OK;
  ret |= pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  ret |= pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  ret |= pthread_mutex_lock(&init_barrier_mutex);
  if(config->state==INITIALIZING_THREAD_STATE
		  && Is_In_Mode(config, current_mode))
	  reached_init_barrier++;
  ret |= pthread_cond_signal(&init_barrier_cond);
#if defined RUNTIME_DEBUG
  printf("Thread %s initialized\n", config->name);
#endif
  config->state=AWAITING_MODE_THREAD_STATE;
  ret |= pthread_cond_wait(config->wait_mode_event, &init_barrier_mutex);
  struct timespec offset_timespec;
  ret |= pthread_mutex_unlock(&init_barrier_mutex);
#if defined RUNTIME_DEBUG
  printf("Thread %s released in current mode\n", config->name);
#endif
  switch (config->protocol)
    {
      case PERIODIC:
      {
        periodic_thread_config_t * per_thr = config->thread;
        per_thr->iteration_counter=0;
        if(per_thr->offset)
        {
        	unsigned long sec_off = per_thr->offset/1000;
        	unsigned long ns_off = (per_thr->offset*1000 - (sec_off * 1000000))*1000;
        	offset_timespec.tv_sec = start_timespec.tv_sec+sec_off;
        	offset_timespec.tv_nsec = start_timespec.tv_nsec+ns_off;
        	while(offset_timespec.tv_nsec >= 1000000000) {
        		offset_timespec.tv_nsec -= 1000000000;
        		offset_timespec.tv_sec++;
        	}
        	while(offset_timespec.tv_nsec < 0) {
        		offset_timespec.tv_nsec += 1000000000;
        		offset_timespec.tv_sec--;
        	}
        	clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &offset_timespec, NULL);
        }
        break;
      }
      default:
    	break;
    }
  return ret;
}

StatusType release_task_in_mode_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;
	ret |= pthread_cond_signal(config->wait_mode_event);
	return ret;
}

StatusType create_thread_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;
	// nothing to do here (start_thread_multiarch also create
	// thread on POSIX
	return ret;
}

StatusType start_thread_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;
	ret |= create_thread(config->priority,
			PTHREAD_STACK_MIN,
			config->compute_entrypoint,
			(pthread_t*) config->thread_id,
			SCHED_FIFO,
			0);
	return ret;
}

StatusType stop_thread_multiarch(thread_config_t * config)
{
	pthread_t * tid = (pthread_t*) config->thread_id;
	return pthread_cancel(*tid);
}

StatusType await_periodic_dispatch_multiarch(periodic_thread_config_t * info)
{
  StatusType ret = E_OK;
  struct timespec periodic_timespec;
  unsigned long sec_per = info->period/1000;
  unsigned long ns_per = (info->period*1000 - (sec_per * 1000000))*1000;

  unsigned long sec_off = info->offset/1000;
  unsigned long ns_off = (info->offset*1000 - (sec_off * 1000000))*1000;

  info->iteration_counter++;
  
  periodic_timespec.tv_sec = start_timespec.tv_sec + info->iteration_counter*sec_per + sec_off;
  periodic_timespec.tv_nsec = start_timespec.tv_nsec + info->iteration_counter*ns_per + ns_off;

  while(periodic_timespec.tv_nsec >= 1000000000) {
    periodic_timespec.tv_nsec -= 1000000000;
    periodic_timespec.tv_sec++;
  }
  while(periodic_timespec.tv_nsec < 0) {
	periodic_timespec.tv_nsec += 1000000000;
	periodic_timespec.tv_sec--;
  }

#ifdef RUNTIME_DEBUG
  printf("Periodic wait until %ld sec %ld ns, with period (msec)=%ld\n",periodic_timespec.tv_sec, periodic_timespec.tv_nsec, info->period);
#endif
  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &periodic_timespec, NULL);
#ifdef RUNTIME_DEBUG
  printf("Periodic dispatch of %s\n", info->parent->name);
#endif
  return ret;
}

void sporadic_thread_sleep(sporadic_thread_config_t * info)
{
  // absolute timeout
  struct timespec sporadic_timespec;
  sporadic_timespec.tv_sec = info->parent->dispatch_time.msec/1000+info->time_interval/1000;
  sporadic_timespec.tv_nsec = info->parent->dispatch_time.nsec + (info->time_interval-((info->time_interval/1000)*1000))*1000000;
  while(sporadic_timespec.tv_nsec >= 1000000000) {
	  sporadic_timespec.tv_nsec -= 1000000000;
	  sporadic_timespec.tv_sec++;
  }
  while(sporadic_timespec.tv_nsec < 0) {
	  sporadic_timespec.tv_nsec += 1000000000;
	  sporadic_timespec.tv_sec--;
  }
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
  printf("Sporadic wait until sec %ld nsec %ld, period (ms) = %ud\n", sporadic_timespec.tv_sec, sporadic_timespec.tv_nsec, info->time_interval);
  printf("Dispatch time was %ud msec %ud nsec\n", info->parent->dispatch_time.msec, info->parent->dispatch_time.nsec);
#endif

  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &sporadic_timespec, NULL);
}

StatusType init_hybrid_timer(hybrid_thread_config_t * info)
{
  clock_gettime(CLOCK_MONOTONIC, info->start_timespec);
  info->nb_period = 0;
  return 0;
}

StatusType hybrid_reset_timer(hybrid_thread_config_t * info)
{
  info->nb_period += 1;
  return E_OK;
}

StatusType timedwait_hybrid(hybrid_thread_config_t * info)
{
  unsigned long sec = info->time_interval/1000;
  unsigned long ns = (info->time_interval*1000 - (sec * 1000000))*1000;

  struct timespec next_period;
  next_period.tv_sec = info->start_timespec->tv_sec + sec * info->nb_period;
  next_period.tv_nsec = info->start_timespec->tv_nsec + ns * info->nb_period;

  while(info->start_timespec->tv_nsec > 1000000000)
    {
  	  info->start_timespec->tv_nsec -= 1000000000;
  	  info->start_timespec->tv_sec++;
    }
  while(info->start_timespec->tv_nsec < 0) {
 	info->start_timespec->tv_nsec += 1000000000;
 	info->start_timespec->tv_sec--;
   }

  int ret = pthread_cond_timedwait(info->parent->global_q->event,
                                   info->parent->global_q->rez,
                                   &next_period);
  return ret;
}

StatusType timedwait_timed_thread(timed_thread_config_t * info)
{
  // relative timeout
  struct timespec timed_timespec;
  clock_gettime(CLOCK_REALTIME, &timed_timespec); // must be CLOCK_REALTIME for pthread_cond_timedwait

  timed_timespec.tv_sec = timed_timespec.tv_sec+info->time_interval/1000;
  unsigned long add = (info->time_interval-((info->time_interval/1000)*1000))*1000000;
  timed_timespec.tv_nsec = timed_timespec.tv_nsec+(info->time_interval-((info->time_interval/1000)*1000))*1000000;
  while(timed_timespec.tv_nsec >= 1000000000) {
	  timed_timespec.tv_nsec -= 1000000000;
	  timed_timespec.tv_sec++;
  }
  while(timed_timespec.tv_nsec < 0) {
	  timed_timespec.tv_nsec += 1000000000;
	  timed_timespec.tv_sec--;
  }
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
  printf("Timed wait until sec %ld nsec %ld, timeout (ms) = %u\n", timed_timespec.tv_sec, timed_timespec.tv_nsec, info->time_interval);
  printf("Dispatch time was %ld msec %ld nsec\n", info->parent->dispatch_time.msec, info->parent->dispatch_time.nsec);
#endif
  int ret = pthread_cond_timedwait(info->parent->global_q->event,
                                   info->parent->global_q->rez,
								   &timed_timespec);
  return ret;
}

StatusType create_thread (unsigned int priority,
		   unsigned int stack_size,
		   void (*start_routine)(void),
		   void * tid,
		   int policy,
		   unsigned int core_id)
{
  StatusType                sys_ret;
  pthread_attr_t     attr;
  struct sched_param param;

  sys_ret = pthread_attr_init (&attr);
  if (sys_ret != 0)
    {
#if defined DEBUG
      printf("Error: thread attr init %d\n", sys_ret);
#endif
      return sys_ret;
    }

  sys_ret = pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
#if defined DEBUG
  if(sys_ret != 0)
  	printf("Error: set inherit sched\n");
#endif

  sys_ret = pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);
  if (sys_ret != 0)
    {
#ifdef DEBUG
      printf("Error: thread attr set scope %d\n", sys_ret);
#endif
      return sys_ret;
    }

  if (stack_size != 0)
    {
  // On some systems, pthread_attr_setstacksize() can fail with the error
  // EINVAL if stacksize is not a multiple of the system page size.
      sys_ret = pthread_attr_setstacksize (&attr, stack_size);
      if (sys_ret != 0)
	{
#ifdef DEBUG
          printf("Error: thread attr set stack size %d\n", sys_ret);
#endif
	  return sys_ret;
	}
    }

  sys_ret = pthread_attr_setschedpolicy(&attr, policy);
#if defined DEBUG
  if(sys_ret != 0)
  {
  	printf("Error: set sched policy\n");
  	return sys_ret;
  }
#endif

  param.sched_priority = priority;
  sys_ret = pthread_attr_setschedparam (&attr, &param);
  if (sys_ret != 0)
    {
#ifdef DEBUG
      printf("Error: thread attr set sched param %d\n", sys_ret);
#endif
      return sys_ret;
    }

  sys_ret = pthread_create ((pthread_t*)tid, &attr, (void* (*)(void*))start_routine, NULL);

  if(sys_ret != 0)
  {
#if defined DEBUG
  	printf("Error: pthread_create\n");
#endif
  	return sys_ret;
  }

  sys_ret = bind_thread_to_core((pthread_t*)tid, core_id);

  return sys_ret;
}

// for aadl_dispatch
int bind_thread_to_core(pthread_t * tid, int core_id) {
  int sys_ret = 0;
  int num_cores = sysconf(_SC_NPROCESSORS_ONLN);
  if (core_id < 0 || core_id > num_cores-1)
  {
#if defined DEBUG
    printf("Error: invalid core_id when binding thread %d\n", core_id);
#endif
	return -1;
  }

  cpu_set_t cpuset;
  CPU_ZERO(&cpuset);
  CPU_SET(core_id, &cpuset);

  sys_ret = pthread_setaffinity_np(*tid, sizeof(cpu_set_t), &cpuset);
#if defined DEBUGS
  if(sys_ret != 0)
  	printf("Error: set affinity failed\n");
#endif
  return sys_ret;
}

StatusType lock_init_barrier()
{
	StatusType ret = E_OK;
	ret = pthread_mutex_lock(&init_barrier_mutex);
	return ret;
}

StatusType unlock_init_barrier()
{
	StatusType ret = E_OK;
	ret = pthread_mutex_unlock(&init_barrier_mutex);
	return ret;
}

StatusType wait_threads_init_multiarch(int threads_under_init)
{
	StatusType ret = E_OK;
	reached_init_barrier = 0;
	while(reached_init_barrier<threads_under_init)
	{
		ret |= pthread_cond_wait(&init_barrier_cond, &init_barrier_mutex);
	}
	return ret;
}

StatusType await_time_triggered_dispatch_multiarch(time_triggered_thread_config_t * config)
{
  //  unsigned int deadline_per_mode_ns = get_start_time_ns() + config->deadline_per_mode[mode]*1000000;
  //  keep_busy_until(deadline_per_mode_ns);
  StatusType ret = pthread_mutex_lock(config->release_rez);
  if(config->released == 0)
    ret |= pthread_cond_wait(config->release_event, config->release_rez);
  config->released=0;
  ret |= pthread_mutex_unlock(config->release_rez);
  return ret;
}

StatusType init_time_triggered_config(time_triggered_thread_config_t * config)
{
  StatusType ret = pthread_cond_init(config->release_event, NULL);
  ret |= pthread_mutex_init(config->release_rez, NULL);
  if (config->parent->global_q != NULL)
    ret |= init_port_list_mutex(config->parent->global_q);
  return ret;
}
