#include "aadl_multiarch.h"

#include "arinc653/sampling.h"
#include "arinc653/queueing.h"
#include "arinc653/semaphore.h"
#include "arinc653/event.h"

#include "main.h"

#ifdef DEBUG
#include <libc/stdio.h>

void print_debug(char * str)
{
  printf(str);
}
#endif

// for aadl_port
StatusType lock_port_reference(thread_input_port_t * port)
{
  StatusType ret = E_OK;
#ifdef RUNTIME_DEBUG
  printf("About to lock %s\n", port->parent->parent->name);
#endif
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  WAIT_SEMAPHORE(*(port->parent->rez),
		  INFINITE_TIME_VALUE,
		  &ret);
#endif
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
#if defined DEBUG
  if(ret!=0)
		printf("ERROR in wait semaphore%d\n", ret);
#endif
  return ret;
}

StatusType unlock_port_reference(thread_input_port_t * port)
{
  StatusType ret = E_OK;
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SIGNAL_SEMAPHORE(*(port->parent->rez), &ret);
#endif
#ifdef RUNTIME_DEBUG
  printf("Unlocked %s, returned %d\n", port->parent->parent->name, ret);
#endif
  return ret;
}

StatusType send_signal(thread_input_port_t * port)
{
  StatusType ret = E_OK;
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SIGNAL_SEMAPHORE(*(port->parent->event), &ret);
#endif
  return ret;
}

StatusType start_thread_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;
	PROCESS_ID_TYPE * proc_id = (PROCESS_ID_TYPE *) config->thread_id;

	switch (config->protocol)
	{
		case PERIODIC:
		{
			periodic_thread_config_t * per_thr = config->thread;
			per_thr->iteration_counter=0;
			if(per_thr->offset)
			{
				SYSTEM_TIME_TYPE offset = (SYSTEM_TIME_TYPE) per_thr->offset;
				DELAYED_START(*proc_id, offset, &ret);
			}
			else
				START(*proc_id, &ret);
			break;
		}
		default:
			START(*proc_id, &ret);
			break;
	}
	return ret;
}

//XXX: must be generated!
extern unsigned int * src_to_dst[];
extern unsigned int nb_dst_per_port[];

extern unsigned int port_to_process[];
extern unsigned int process_to_node[];

StatusType send_out_process(port_reference_t * source,
		void * value,
		unsigned int msg_size)
{
	StatusType ret = E_OK;
	process_port_t * src_port = (process_port_t *) source->directed_port;

	unsigned int src_process, src_node,dst_process, dst_node;
	src_process = port_to_process[src_port->identifier];
	src_node = process_to_node[src_process];

	unsigned int nb_dest = nb_dst_per_port[src_port->identifier];
	unsigned int * dst_port_id_array = src_to_dst[src_port->identifier];
	unsigned int i, already_sent_locally=0;

	for(i=0; i<nb_dest; i++)
	{
		dst_process = port_to_process[dst_port_id_array[i]];
		dst_node = process_to_node[dst_process];
		if(dst_node==src_node) {
			already_sent_locally = 1;
			RETURN_CODE_TYPE arinc_call_ret;
			if(source->type==DATA)
			{

#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SAMPLING)
				// sampling
				SAMPLING_PORT_ID_TYPE * sampling_port = (SAMPLING_PORT_ID_TYPE *) src_port->sys_port;
				WRITE_SAMPLING_MESSAGE(*sampling_port,
						(MESSAGE_ADDR_TYPE) value,
						(MESSAGE_SIZE_TYPE) msg_size,
						&arinc_call_ret);
#ifdef DEBUG
				printf("Value returned by WRITE_SAMPLING_MESSAGE %d\n", arinc_call_ret);
#endif //DEBUG
#endif
			}
			else
			{
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_QUEUEING)
				// queueing
				QUEUING_PORT_ID_TYPE * queuing_port = (QUEUING_PORT_ID_TYPE *) src_port->sys_port;
				SEND_QUEUING_MESSAGE(*queuing_port,
						(MESSAGE_ADDR_TYPE) value,
						(MESSAGE_SIZE_TYPE) msg_size,
						INFINITE_TIME_VALUE,
						&arinc_call_ret
				);
#ifdef DEBUG
				printf("Value returned but SEND_QUEUING_MESSAGE %d\n", arinc_call_ret);
#endif //DEBUG
#endif // POK_NEEDS_ARINC653_QUEUING
			}
		}
		else
		{
			// XXX: nothing to do (inter-node communications handled elsewhere)?
			// or implement protocol specific calls (inter-node communications NOT handled elsewhere) ?
		}

	}
	return ret;
}

StatusType receive_in_process(port_reference_t * destination,
		void * received_msg)
{
	StatusType ret = E_OK;
	process_port_t * dst_port = (process_port_t *) destination->directed_port;

	unsigned int msg_size = destination->msg_size;

	MESSAGE_SIZE_TYPE read_data_size=0;
	if(destination->type==DATA)
	{
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SAMPLING)
		// sampling
		VALIDITY_TYPE is_valid;
		SAMPLING_PORT_ID_TYPE * sampling_port = (SAMPLING_PORT_ID_TYPE *) dst_port->sys_port;
		READ_SAMPLING_MESSAGE(*sampling_port,
				(MESSAGE_ADDR_TYPE) received_msg,
				&read_data_size,
				&is_valid,
				&ret);
#endif // POK_NEEDS_ARINC653_SAMPLING
	}
	else
	{
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_QUEUEING)
		// queueing
		read_data_size = msg_size;
		QUEUING_PORT_ID_TYPE * queuing_port = (QUEUING_PORT_ID_TYPE *) dst_port->sys_port;
		RECEIVE_QUEUING_MESSAGE(*queuing_port,
				INFINITE_TIME_VALUE,
				(MESSAGE_ADDR_TYPE) received_msg,
				&read_data_size,
				&ret
		);
#ifdef DEBUG
			printf("Value returned by RECEIVE_QUEUEING_MESSAGE = %d\n", ret);
#endif //DEBUG
#endif // POK_NEEDS_ARINC653_QUEUING
	}

	return ret;
}

StatusType lock_data(lock_access_protocol_t * protocol)
{
  StatusType ret = 0;
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  WAIT_SEMAPHORE((*protocol->rez),
  		  INFINITE_TIME_VALUE,
  		  &ret);
#endif
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
  return ret;
}

StatusType unlock_data(lock_access_protocol_t * protocol)
{
  StatusType ret = 0;
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SIGNAL_SEMAPHORE((*protocol->rez), &ret);
#endif
  return ret;
}

StatusType icpp_lock_data(icpp_access_protocol_t * protocol)
{ // TODO: add debug messages
  StatusType ret = 0;
  PRIORITY_TYPE ceiling_priority = protocol->ceiling_priority;

  if(ceiling_priority==0)
	return -1;

  PROCESS_ID_TYPE me;
  GET_MY_ID(&me, &ret);
  if(ret!=0)
	return ret;
  PROCESS_STATUS_TYPE status;
  GET_PROCESS_STATUS(me, &status, &ret);
  if(ret!=0)
  	return ret;

  PRIORITY_TYPE base_priority = status.ATTRIBUTES.BASE_PRIORITY;

  SET_PRIORITY(me, ceiling_priority, &ret);
  if(ret!=0)
	return -1;
  if(protocol->base_priority>0)
	return -1;
  protocol->base_priority = base_priority;
  return ret;
}

StatusType icpp_unlock_data(icpp_access_protocol_t * protocol)
{
  StatusType ret = 0;
  PROCESS_ID_TYPE me;
  GET_MY_ID(&me, &ret);
  if(ret!=0)
  	return ret;

  PRIORITY_TYPE base_priority = protocol->base_priority;
  protocol->base_priority = 0;
  SET_PRIORITY(me, base_priority, &ret);
}

int init_port_list_mutex(port_list_t * port_list)
{
  return 0;
}

StatusType lock_port_list(port_list_t * port_list)
{
  StatusType ret = 0;
#ifdef RUNTIME_DEBUG
  printf("About to lock %s\n", port_list->parent->name);
#endif
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  WAIT_SEMAPHORE(*(port_list->rez),
		  INFINITE_TIME_VALUE,
		  &ret);
#endif
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
  return ret;
}

StatusType unlock_port_list(port_list_t * port_list)
{
  StatusType ret = 0;
  // DO NOT USE printf inside critical sections: it locks
  // the std output file and may provoke deadlocks
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  SIGNAL_SEMAPHORE(*(port_list->rez), &ret);
#endif
#ifdef RUNTIME_DEBUG
  printf("Unlocked %s, returned %d\n", port_list->parent->name, ret);
#endif
  return ret;
}

StatusType wait_port_list(port_list_t * port_list)
{
  int ret=0;
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  unlock_port_list(port_list);
  WAIT_SEMAPHORE ((*port_list->event),
		  	  INFINITE_TIME_VALUE,
			  &ret);
  lock_port_list(port_list);
#endif
  return ret;
}

